# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log as logging

from tobiko.shell import custom_script
from tobiko.shell import files
from tobiko.shell import ssh


LOG = logging.getLogger(__name__)

TIMEOUT = 2  # seconds
LOG_FILE_NAME = "dhcp_ping.log"
DHCP_PING_SCRIPT_NAME = "tobiko_dhcp_ping.sh"
DHCP_PING_SCRIPT = """
#!/bin/bash

HOSTNAME_CMD=$(which hostname)
DATE_CMD=$(which date)
NMAP_CMD="sudo $(which nmap) --script broadcast-dhcp-discover 2>/dev/null"

output_file=$1;

rm $output_file;

# First get IP address(es) of the machine
server_ips=$($HOSTNAME_CMD --all-ip-addresses)

while true; do
    current_time=$(/usr/bin/date +"{time_format}");
    nmap_response=$($NMAP_CMD | grep -i "ip offered" | cut -d ":" -f 2 | tr -d " ")


    if [ -n "$nmap_response" ] && [[ $server_ips = *$nmap_response* ]]; then
            response="{result_ok}";
    else
            response="{result_failed}";
    fi;
    echo "{output_result_line}" >> $output_file;
    sleep {timeout};
done;
""".format(  # noqa: E501
    time_format=custom_script.LOG_TIME_FORMAT,
    result_ok=custom_script.RESULT_OK,
    result_failed=custom_script.RESULT_FAILED,
    output_result_line=custom_script.LOG_RESULT_FORMAT,
    timeout=TIMEOUT,
)


def _ensure_script_is_on_server(ssh_client: ssh.SSHClientType) -> None:
    custom_script.ensure_script_is_on_server(
        DHCP_PING_SCRIPT_NAME,
        DHCP_PING_SCRIPT,
        ssh_client=ssh_client)


def _get_script_command(ssh_client: ssh.SSHClientType) -> str:
    homedir = files.get_homedir(ssh_client)
    logfile_path = _get_logfile_path(ssh_client)
    return f"bash {homedir}/{DHCP_PING_SCRIPT_NAME} {logfile_path}"


def _get_log_dir(ssh_client: ssh.SSHClientType = None) -> str:
    return custom_script.get_log_dir(
        "tobiko_dhcp_ping_results", ssh_client)


def _get_logfile_path(ssh_client: ssh.SSHClientType) -> str:
    return f"{_get_log_dir(ssh_client)}/{LOG_FILE_NAME}"


def _get_dhcp_ping_pid(
        ssh_client: ssh.SSHClientType) -> typing.Union[int, None]:
    processes = custom_script.get_process_pid(
        command_line=_get_script_command(ssh_client),
        ssh_client=ssh_client)
    if not processes:
        LOG.debug('no DHCP ping script found.')
    return processes


def _get_dhcp_ping_pids(
        ssh_client: ssh.SSHClientType) -> list:
    pids = custom_script.get_process_pids(
        command_line=_get_script_command(ssh_client),
        ssh_client=ssh_client)
    if not pids:
        LOG.debug('no DHCP ping script found.')
    return pids


def start_dhcp_ping_process(ssh_client: ssh.SSHClientType) -> None:
    _ensure_script_is_on_server(ssh_client)
    if dhcp_ping_process_alive(ssh_client):
        return
    custom_script.start_script(
        _get_script_command(ssh_client),
        ssh_client=ssh_client)


def stop_dhcp_ping_process(ssh_client: ssh.SSHClientType) -> None:
    pids = _get_dhcp_ping_pids(ssh_client)
    for pid in pids:
        custom_script.stop_script(pid, ssh_client=ssh_client)


def dhcp_ping_process_alive(ssh_client: ssh.SSHClientType) -> bool:
    return bool(_get_dhcp_ping_pid(ssh_client))


def check_dhcp_ping_results(ssh_client: ssh.SSHClientType) -> None:
    # Source log file is on the guest vm so ssh_client needs to be used
    # to get it
    src_logfile = _get_logfile_path(ssh_client)
    # Destination is local to where Tobiko is running so no need to pass
    # ssh_client to get_log_dir() function this time
    dest_logfile = f"{_get_log_dir()}/{LOG_FILE_NAME}"
    custom_script.copy_log_file(src_logfile, dest_logfile, ssh_client)

    # Looking for the files locally so no need to pass ssh_client to the
    # _get_log_dir function
    logfiles = custom_script.get_log_files(
        glob_log_pattern=f"{_get_log_dir()}/{LOG_FILE_NAME}")

    custom_script.check_results(logfiles)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import glob
import io
import typing

from oslo_log import log as logging
from oslo_serialization import jsonutils

import tobiko
from tobiko import config
from tobiko.shell import files
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.shell.custom_script import _constants


CONF = config.CONF
LOG = logging.getLogger(__name__)


def ensure_script_is_on_server(
        script_name: str,
        script: str,
        ssh_client: ssh.SSHClientType = None):
    homedir = files.get_homedir(ssh_client)
    sh.execute(
        f"echo '{script}' > {homedir}/{script_name}",
        ssh_client=ssh_client)


def get_log_dir(
        path: str,
        ssh_client: ssh.SSHClientType = None) -> str:
    return files.get_home_absolute_filepath(path, ssh_client)


def get_log_files(glob_log_pattern: str) -> typing.List[str]:
    """return a list of files matching : the pattern"""
    log_files = []
    for filename in glob.glob(glob_log_pattern):
        LOG.info(f'found following log file {filename}')
        log_files.append(filename)
    return log_files


def copy_log_file(
            src_logfile: str,
            dest_logfile: str,
            ssh_client: ssh.SSHClientType) -> None:
    for attempt in tobiko.retry(timeout=60, interval=5):
        # download the http ping results file to local
        try:
            sh.get_file(src_logfile, dest_logfile, ssh_client)
            # Remote log file can be now deleted so that logs from now will be
            # in the "clean" file if they will need to be checked later
            sh.execute(f'rm -f {src_logfile}', ssh_client=ssh_client)
            return
        except sh.ShellCommandFailed as err:
            message = f'Failed to download log file. Error {err}'
            if attempt.is_last:
                tobiko.fail(message)
            else:
                LOG.debug(message)
                LOG.debug('Retrying to download log file...')
                continue
        if attempt.is_last:
            tobiko.fail('Failed to download log file.')


def get_process_pid(
        command_line: str,
        ssh_client: ssh.SSHClientType = None
        ) -> typing.Union[int, None]:
    processes = sh.list_processes(
        command_line=command_line,
        ssh_client=ssh_client)
    if processes:
        return processes.first.pid
    return None


def get_process_pids(
        command_line: str,
        ssh_client: ssh.SSHClientType = None,
        ) -> list:
    processes = sh.list_processes(
        command_line=command_line,
        ssh_client=ssh_client)
    pids = []
    if processes:
        pids = [p.pid for p in processes]
    return pids


def check_results(
        log_filenames: typing.List[str]):

    failure_limit = CONF.tobiko.rhosp.max_ping_loss_allowed
    for filename in log_filenames:
        with io.open(filename, 'rt') as fd:
            LOG.info(f'checking custom script log file: {filename}, '
                     f'failure_limit is :{failure_limit}')
            failures_list = []
            for log_line in fd.readlines():
                log_line_json = jsonutils.loads(log_line.rstrip())
                if log_line_json['response'] != _constants.RESULT_OK:
                    # NOTE(salweq): Add file name to the failure line
                    #               just for the debugging purpose
                    log_line_json['filename'] = filename
                    failures_list.append(log_line_json)

            failures_len = len(failures_list)
            if failures_len > 0:
                failures_str = '\n'.join(
                    [str(failure) for failure in failures_list])
                LOG.warning(f'found custom script failures:\n{failures_str}')
            else:
                LOG.debug(f'no failures in custom script log file: {filename}')

            tobiko.truncate_logfile(filename)

            if failures_len >= failure_limit:
                tobiko.fail(f'{failures_len} failures found '
                            f'in file: {failures_list[-1]["filename"]}')


def start_script(
        command_line: str,
        ssh_client: ssh.SSHClientType = None):
    process = sh.process(
        command_line, ssh_client=ssh_client)
    process.execute()


def stop_script(
        pid: int,
        ssh_client: ssh.SSHClientType = None):
    # expect_exit_status=None -> will not fail if the kill command fails, maybe
    # the process had already died
    sh.execute(f'kill {pid}',
               ssh_client=ssh_client,
               sudo=True,
               expect_exit_status=None)
    # wait until http ping process disappears
    sh.wait_for_processes(timeout=120,
                          sleep_interval=5,
                          ssh_client=ssh_client,
                          pid=pid)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import glob
import json
import io
import os
import time
import typing


import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.shell import files
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.shell.ping import _interface
from tobiko.shell.ping import _exception
from tobiko.shell.ping import _parameters
from tobiko.shell.ping import _statistics


CONF = config.CONF
LOG = log.getLogger(__name__)


TRANSMITTED = 'transmitted'
DELIVERED = 'delivered'
UNDELIVERED = 'undelivered'
RECEIVED = 'received'
UNRECEIVED = 'unreceived'


PingHostType = typing.Union['str', netaddr.IPAddress]


def list_reachable_hosts(hosts: typing.Iterable[PingHostType],
                         **params) -> tobiko.Selection[PingHostType]:
    reachable_host, _ = ping_hosts(hosts, **params)
    return reachable_host


def list_unreachable_hosts(hosts: typing.Iterable[PingHostType],
                           **params) -> tobiko.Selection[PingHostType]:
    _, unreachable_host = ping_hosts(hosts, **params)
    return unreachable_host


PingHostsResultType = typing.Tuple[tobiko.Selection[PingHostType],
                                   tobiko.Selection[PingHostType]]


def wait_for_ping_hosts(hosts: typing.Iterable[PingHostType],
                        check_unreachable=False,
                        retry_count: int = None,
                        retry_timeout: tobiko.Seconds = None,
                        retry_interval: tobiko.Seconds = None,
                        **params) \
        -> None:
    if retry_timeout is None:
        retry_timeout = params.get('timeout')
    LOG.debug("Wait for ping hosts:\n"
              f"  hosts: {hosts}\n"
              f"  check_unreachable: {check_unreachable}\n"
              f"  retry_count: {retry_count}\n"
              f"  retry_timeout: {retry_timeout}\n"
              f"  retry_interval: {retry_interval}\n"
              f"  **params: {params}\n")
    for attempt in tobiko.retry(count=retry_count,
                                timeout=retry_timeout,
                                interval=retry_interval,
                                default_timeout=300.,
                                default_interval=5.):
        reachable, unreachable = ping_hosts(hosts, **params)
        if check_unreachable:
            hosts = reachable
        else:
            hosts = unreachable
        if hosts:
            if attempt.is_last:
                if check_unreachable:
                    raise _exception.ReachableHostsException(
                        hosts=hosts,
                        timeout=attempt.timeout,
                        elapsed_time=attempt.elapsed_time) from None
                else:
                    raise _exception.UnreachableHostsException(
                        hosts=hosts,
                        timeout=attempt.timeout,
                        elapsed_time=attempt.elapsed_time) from None
        else:
            break

    else:
        raise RuntimeError('Broken retry loop')  # This is a bug


def ping_hosts(hosts: typing.Iterable[PingHostType],
               count: typing.Optional[int] = None,
               **params) -> PingHostsResultType:
    if count is None:
        count = 1
    else:
        count = int(count)
    reachable = tobiko.Selection[PingHostType]()
    unreachable = tobiko.Selection[PingHostType]()
    for host in hosts:
        try:
            result = ping(host, count=count, **params)
        except _exception.PingError:
            LOG.exception('Error pinging host: %r', host)
            unreachable.append(host)
        else:
            if result.received:
                reachable.append(host)
            else:
                unreachable.append(host)
    return reachable, unreachable


def ping(host: PingHostType, until=TRANSMITTED, check: bool = True,
         **ping_params) -> _statistics.PingStatistics:
    """Send ICMP messages to host address until timeout

    :param host: destination host address
    :param ping_params: parameters to be forwarded to :mod:`get_statistics`
        function
    :returns: PingStatistics
    """
    return get_statistics(host=host, until=until, check=check, **ping_params)


def ping_until_delivered(host, **ping_params):
    """Send 'count' ICMP messages

    Send 'count' ICMP messages

    ICMP messages are considered delivered when they have been
    transmitted without being counted as errors.

    :param host: destination host address
    :param ping_params: parameters to be forwarded to :mod:`get_statistics`
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before delivering all
        expected count messages
    """
    return ping(host=host, until=DELIVERED, **ping_params)


def ping_until_undelivered(host, **ping_params):
    """Send ICMP messages until it fails to deliver messages

    Send ICMP messages until it fails to deliver 'count' messages

    ICMP messages are considered undelivered when they have been
    transmitted and they have been counted as error in ping statistics (for
    example because of errors into the route to remote address).

    :param host: destination host address
    :param ping_params: parameters to be forwarded to :mod:`get_statistics`
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before failing delivering
        expected 'count' of messages
    """
    return ping(host=host, until=UNDELIVERED, **ping_params)


def ping_until_received(host, **ping_params):
    """Send ICMP messages until it receives messages back

    Send ICMP messages until it receives 'count' messages back

    ICMP messages are considered received when they have been
    transmitted without any routing errors and they are received back

    :param host: destination host address
    :param ping_params: parameters to be forwarded to :mod:`get_statistics`
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before receiving all
        expected 'count' of messages
    """
    return ping(host=host, until=RECEIVED, **ping_params)


def ping_until_unreceived(host, **ping_params):
    """Send ICMP messages until it fails to receive messages

    Send ICMP messages until it fails to receive 'count' messages back.

    ICMP messages are considered unreceived when they have been
    transmitted without any routing error but they failed to be received
    back (for example because of network filtering).

    :param host: destination host address
    :param ping_params: parameters to be forwarded to :mod:`get_statistics`
        function
    :returns: PingStatistics
    :raises: PingFailed in case timeout expires before failed receiving
        expected 'count' of messages
    """
    return ping(host=host, until=UNRECEIVED, **ping_params)


def get_statistics(parameters=None, ssh_client=None, until=None, check=True,
                   **ping_params) -> _statistics.PingStatistics:
    parameters = _parameters.get_ping_parameters(default=parameters,
                                                 **ping_params)
    statistics = _statistics.PingStatistics()
    for partial_statistics in iter_statistics(parameters=parameters,
                                              ssh_client=ssh_client,
                                              until=until, check=check):
        statistics += partial_statistics
        LOG.debug("%r", statistics)

    return statistics


def iter_statistics(parameters=None, ssh_client=None, until=None, check=True,
                    **ping_params):
    parameters = _parameters.get_ping_parameters(default=parameters,
                                                 **ping_params)
    now = time.time()
    end_of_time = now + parameters.timeout
    deadline = parameters.deadline
    transmitted = 0
    received = 0
    undelivered = 0
    count = 0
    enlapsed_time = None

    while deadline > 0. and count < parameters.count:
        if enlapsed_time is not None and enlapsed_time < deadline:
            # Avoid busy waiting when errors happens
            sleep_time = deadline - enlapsed_time
            LOG.debug('Waiting %s seconds before next ping execution',
                      sleep_time)
            time.sleep(sleep_time)

        start_time = time.time()

        # splitting total timeout interval into smaller deadline intervals will
        # cause ping command to be executed more times allowing to handle
        # temporary packets routing problems
        if until == RECEIVED:
            execute_parameters = _parameters.get_ping_parameters(
                default=parameters,
                deadline=deadline,
                count=(parameters.count - count),
                timeout=end_of_time - now)
        else:
            # Deadline ping parameter cause ping to be executed until count
            # messages are received or deadline is expired
            # Therefore to count messages not of received type we have to
            # simulate deadline parameter limiting the maximum number of
            # transmitted messages
            execute_parameters = _parameters.get_ping_parameters(
                default=parameters,
                deadline=deadline,
                count=min(parameters.count - count,
                          parameters.interval * deadline),
                timeout=end_of_time - now)

        # Command timeout would typically give ping command additional seconds
        # to safely reach deadline before shell command timeout expires, while
        # in the same time adding an extra verification to forbid using more
        # time than expected considering the time required to make SSH
        # connection and running a remote shell
        output = execute_ping(parameters=execute_parameters,
                              ssh_client=ssh_client,
                              check=check)

        if output:
            statistics = _statistics.parse_ping_statistics(
                output=output, begin_interval=now,
                end_interval=time.time())

            yield statistics

            transmitted += statistics.transmitted
            received += statistics.received
            undelivered += statistics.undelivered
        else:
            # Assume 1 transmitted undelivered package when unable to get
            # ping output
            transmitted += 1
            undelivered += 1

        count = {None: 0,
                 TRANSMITTED: transmitted,
                 DELIVERED: transmitted - undelivered,
                 UNDELIVERED: undelivered,
                 RECEIVED: received,
                 UNRECEIVED: transmitted - received}[until]

        now = time.time()
        deadline = min(int(end_of_time - now), parameters.deadline)
        enlapsed_time = now - start_time
        if enlapsed_time > 0.:
            LOG.debug('Ping execution took %s seconds', enlapsed_time)

    if until and count < parameters.count:
        raise _exception.PingFailed(count=count,
                                    expected_count=parameters.count,
                                    timeout=parameters.timeout,
                                    message_type=until)


def execute_ping(parameters, ssh_client=None, check=True):
    command = _interface.get_ping_command(parameters=parameters,
                                          ssh_client=ssh_client)

    try:
        result = sh.execute(command=command,
                            ssh_client=ssh_client,
                            timeout=parameters.deadline + 3.,
                            expect_exit_status=None,
                            network_namespace=parameters.network_namespace)
    except (sh.ShellError, tobiko.RetryLimitError) as ex:
        LOG.exception("Error executing ping command")
        stdout = ex.stdout if hasattr(ex, "stdout") else None
        stderr = ex.stderr if hasattr(ex, "stderr") else None
    else:
        stdout = result.stdout
        stderr = result.stderr

    if stdout:
        output = str(stdout)
    else:
        output = None

    if stderr:
        error = str(stderr)
        if check and result.exit_status:
            handle_ping_command_error(error=error)

    return output


def handle_ping_command_error(error):
    for error in error.splitlines():
        error = error.strip()
        if error:
            prefix = 'ping: '
            if error.startswith('ping: '):
                error = error[len(prefix):]
            handle_ping_bad_address_error(error)
            handle_ping_local_error(error)
            handle_ping_connect_error(error)
            handle_ping_send_to_error(error)
            handle_ping_unknow_host_error(error)
            raise _exception.PingError(details=error)


def handle_ping_bad_address_error(text):
    prefix = 'bad address '
    if text.startswith(prefix):
        address = text[len(prefix):].replace("'", '').strip()
        raise _exception.BadAddressPingError(address=address)


def handle_ping_local_error(text):
    prefix = 'local error: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.LocalPingError(details=details)


def handle_ping_connect_error(text):
    prefix = 'connect: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.ConnectPingError(details=details)


def handle_ping_send_to_error(text):
    prefix = 'sendto: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.SendToPingError(details=details)


def handle_ping_unknow_host_error(text):
    prefix = 'unknown host'
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.UnknowHostError(details=details)

    prefix = 'unreachable-host: '
    if text.startswith(prefix):
        details = text[len(prefix):].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': Name or service not known'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': No route to host'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip()
        raise _exception.UnknowHostError(details=details)

    suffix = ': Unknown host'
    if text.endswith(suffix):
        details = text[:-len(suffix)].strip().split()[-1]
        raise _exception.UnknowHostError(details=details)


def ping_to_json(ping_result: _statistics.PingStatistics) -> str:
    '''Transform an iter_statistics.statistics object
    into a json string with ping ip and result'''
    destination = str(ping_result.destination)
    transmitted = ping_result.transmitted
    received = ping_result.received
    timestamp = time.ctime(ping_result.begin_interval)
    ping_result_line_dict = {"destination": destination,
                             "transmitted": transmitted,
                             "received": received,
                             "timestamp": timestamp}
    return json.dumps(ping_result_line_dict)


def write_ping_to_file(ping_ip=None, output_dir='tobiko_ping_results',
                       interval=None):
    '''use iter_statistics to ping a host and record statistics
    put results in output_dir filenames correlate with vm fip'''
    output_dir_path = f'{sh.get_user_home_dir()}/{output_dir}'
    if not os.path.exists(output_dir_path):
        os.makedirs(output_dir_path)
    output_filename = f'ping_{ping_ip}.log'
    output_path = os.path.join(output_dir_path, output_filename)
    LOG.info(f'starting ping process to > {ping_ip} , '
             f'output file is : {output_path}')
    ping_result_statistics = iter_statistics(parameters=None,
                                             host=ping_ip, until=None,
                                             timeout=99999,
                                             interval=interval,
                                             check=True)
    for ping_result in ping_result_statistics:
        with open(output_path, "at") as ping_result_file:
            ping_result_file.write(ping_to_json(ping_result) + "\n")
            time.sleep(5)


def get_vm_ping_log_files(glob_ping_log_pattern='tobiko_ping_results/ping_'
                                                '*.log'):
    """return a list of files mathcing : the pattern"""
    glob_path = f'{sh.get_user_home_dir()}/{glob_ping_log_pattern}'
    for filename in glob.glob(glob_path):
        LOG.info(f'found following ping_vm_log files {filename}')
        vm_ping_log_filename = filename
        yield vm_ping_log_filename


def check_ping_statistics():
    """Gets a list of ping_vm_log files and
    iterates their lines, checks if max ping
    failures have been reached per fip=file"""
    failure_limit = CONF.tobiko.rhosp.max_ping_loss_allowed
    ping_files_found = False
    # iterate over ping_vm_log files:
    for filename in list(get_vm_ping_log_files()):
        ping_files_found = True
        with io.open(filename, 'rt') as fd:
            LOG.info(f'checking ping log file: {filename}, '
                     f'failure_limit is :{failure_limit}')
            ping_failures_list = []
            for ping_line in fd.readlines():
                ping_line = json.loads(ping_line.rstrip())
                if ping_line['transmitted'] != ping_line['received']:
                    ping_failures_list.append(ping_line)

            ping_failures_len = len(ping_failures_list)
            if ping_failures_len > 0:
                ping_failures_str = '\n'.join(
                    [str(ping_failure) for ping_failure in ping_failures_list])
                LOG.warning(f'found ping failures:\n{ping_failures_str}')
            else:
                LOG.info(f'no failures in ping log file: {filename}')

            tobiko.truncate_logfile(filename)

            if ping_failures_len >= failure_limit:
                tobiko.fail(f'{ping_failures_len} pings failure found '
                            f'to vm fip destination: '
                            f'{ping_failures_list[-1]["destination"]}')

    if not ping_files_found:
        tobiko.fail('No ping log files found')


def skip_check_ping_statistics():
    for filename in list(get_vm_ping_log_files()):
        tobiko.truncate_logfile(filename)
        LOG.info(f'skipping ping failures in ping log file: {filename}')


def _get_ping_pid(
        address: typing.Union[str, netaddr.IPAddress, None] = None,
        ssh_client: ssh.SSHClientType = None) -> typing.Union[int, None]:
    ping_command = 'ping'
    if address is not None:
        ping_command += f' .*{address}'
    ping_processes = sh.list_processes(command_line=ping_command,
                                       ssh_client=ssh_client)
    if not ping_processes:
        LOG.debug('no ping processes were found')
        return None
    else:
        return ping_processes.unique.pid


def ping_alive(address: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
               ssh_client: ssh.SSHClientType = None,
               **kwargs) -> bool:
    return bool(_get_ping_pid(address=address, ssh_client=ssh_client))


def stop_ping(address: typing.Union[str, netaddr.IPAddress],
              ssh_client: ssh.SSHClientType = None,
              **kwargs):  # noqa; pylint: disable=W0613
    pid = _get_ping_pid(address=address, ssh_client=ssh_client)
    if pid:
        LOG.info(f'ping process to > {address} already running '
                 f'with PID: {pid}')
        # the SIGINT signal makes ping write the "ping statistics" block
        # before exiting
        sh.execute(f'kill -s SIGINT {pid}', ssh_client=ssh_client, sudo=True)


def _get_ping_logs_filepath(address: typing.Union[str, netaddr.IPAddress],
                            path: str,
                            ssh_client: ssh.SSHClientType = None) -> str:
    final_dir = files.get_home_absolute_filepath(path, ssh_client)
    filename = f'ping_{address}.log'
    return os.path.join(final_dir, filename)


# TODO(eolivare): replace check_ping_statistics with check_ping_results
def check_ping_results(address: typing.Union[str, netaddr.IPAddress],
                       output_dir: str = 'tobiko_ping_results',
                       ssh_client: ssh.SSHClientType = None,
                       **kwargs):  # noqa; pylint: disable=W0613
    testcase = tobiko.get_test_case()
    testcase.assertFalse(ping_alive(address, ssh_client))

    # This function expects that the result file is available locally already
    logfile = _get_ping_logs_filepath(address, output_dir, ssh_client)
    try:
        ping_log_raw = sh.execute(
            f"cat {logfile}", ssh_client=ssh_client).stdout
    except sh.ShellCommandFailed as err:
        if config.is_prevent_create():
            # Tobiko is not expected to create resources in this run
            # so ping should be already running and log file should
            # be already there, if it is not, it should fail
            tobiko.fail('Failed to read ping log from the file. '
                        f'Ping Destination IP address: {address}; '
                        f'Logfile: {logfile}')
        else:
            # Tobiko is creating resources so it is normal that file was not
            # there yet
            LOG.debug(f'Failed to read ping log from the file. '
                      f'Error: {err}')
            return

    LOG.debug(f'ping log raw: {ping_log_raw}')
    if not ping_log_raw:
        if config.is_prevent_create():
            # Tobiko is not expected to create resources in this run
            # so ping should be already running and log file should
            # be already there, if it is not, it should fail
            tobiko.fail('Failed empty ping file.')
        else:
            LOG.debug('Failed ping log file empty')
            return

    files.truncate_client_logfile(logfile, ssh_client)

    ping_stats = _statistics.parse_ping_statistics(ping_log_raw)

    testcase.assertGreater(ping_stats.transmitted, 0)
    testcase.assertGreater(ping_stats.received, 0)
    testcase.assertLessEqual(ping_stats.transmitted - ping_stats.received,
                             CONF.tobiko.rhosp.max_ping_loss_allowed)


def start_background_ping(address: typing.Union[str, netaddr.IPAddress],
                          output_path: str,
                          ssh_client: ssh.SSHClientType = None):
    parameters = _parameters.get_ping_parameters(host=address,
                                                 count=0,
                                                 deadline=0)
    command = _interface.get_ping_command(parameters, ssh_client)
    # both stdout and stderr need to be written to the provided log file
    command += '2>&1'
    command += f'> {output_path}'
    process = sh.process(command, ssh_client=ssh_client)
    process.execute()


# TODO(eolivare): replace write_ping_to_file with execute_ping_in_background
def execute_ping_in_background(address: typing.Union[str, netaddr.IPAddress],
                               output_dir: str = 'tobiko_ping_results',
                               ssh_client: ssh.SSHClientType = None,
                               **kwargs):  # noqa; pylint: disable=W0613
    output_path = _get_ping_logs_filepath(address, output_dir, ssh_client)
    LOG.info(f'starting ping process to > {address} , '
             f'output file is : {output_path}')
    # just in case there is some leftover file from previous run,
    # it needs to be removed, otherwise ping will append new log
    # to the end of the existing file and this will make output
    # file to be malformed
    files.remove_old_logfile(output_path, ssh_client=ssh_client)

    # Stop ping in case it is running
    stop_ping(address, ssh_client)

    # Start ping again
    start_background_ping(address, output_path, ssh_client)

    # if ping does not start properly, fail the test
    for attempt in tobiko.retry(count=5, sleep_time=.5):
        if ping_alive(address, ssh_client):
            return
        elif attempt.is_last:
            tobiko.fail('background ping process did not start')



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.iperf3 import _assert
from tobiko.shell.iperf3 import _execute
from tobiko.shell.iperf3 import _interface
from tobiko.shell.iperf3 import _parameters


assert_has_bandwith_limits = _assert.assert_has_bandwith_limits
execute_iperf3_client_in_background = \
        _execute.execute_iperf3_client_in_background
check_iperf3_client_results = _execute.check_iperf3_client_results
get_iperf3_logs_filepath = _execute.get_iperf3_logs_filepath
iperf3_client_alive = _execute.iperf3_client_alive
stop_iperf3_client = _execute.stop_iperf3_client
start_iperf3_server = _execute.start_iperf3_server
parse_json_stream_output = _execute.parse_json_stream_output

get_iperf3_client_command = _interface.get_iperf3_client_command

Iperf3ClientParameters = _parameters.Iperf3ClientParameters
iperf3_client_parameters = _parameters.iperf3_client_parameters



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import metalsmith
import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.podified import _openshift
from tobiko import rhosp
from tobiko.shell import files
from tobiko.shell import iperf3
from tobiko.shell import http_ping
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import _overcloud
from tobiko.tripleo import _undercloud
from tobiko.tripleo import containers
from tobiko.tripleo import nova as tripleo_nova

CONF = config.CONF
LOG = log.getLogger(__name__)


class TripleoTopology(rhosp.RhospTopology):

    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'tripleo_neutron_dhcp',
        neutron.L3_AGENT:  'tripleo_neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'tripleo_neutron_ovs_agent',
        neutron.METADATA_AGENT: 'tripleo_neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'tripleo_ovn_controller',
        neutron.OVN_BGP_AGENT: 'tripleo_ovn_bgp_agent',
        neutron.FRR: 'tripleo_frr',
        neutron.NEUTRON: 'tripleo_neutron_api'
    }

    agent_to_container_name_mappings = {
        neutron.DHCP_AGENT: 'neutron_dhcp',
        neutron.L3_AGENT:  'neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'neutron_ovs_agent',
        neutron.METADATA_AGENT: 'neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'ovn_controller',
        neutron.OVN_BGP_AGENT: 'ovn_bgp_agent',
        neutron.FRR: 'frr'
    }

    config_file_mappings = {
        'ml2_conf.ini': '/var/lib/config-data/puppet-generated/neutron'
                        '/etc/neutron/plugins/ml2/ml2_conf.ini',
        'bgp-agent.conf': '/var/lib/config-data/ansible-generated/'
                          'ovn-bgp-agent/etc/ovn-bgp-agent/bgp-agent.conf'
    }

    # TODO: add more known subgrups here
    known_subgroups: typing.List[str] = ['controller', 'compute']

    # In TripleO we need to parse log files directly
    file_digger_class = files.LogFileDigger

    # This is dict which handles mapping of the log file and systemd_unit (if
    # needed) for the OpenStack services
    # Format of this dict is like below:
    # service_name: (log_filename, systemd_unit_name)
    log_names_mappings = {
        neutron.SERVER: '/var/log/containers/neutron/server.log*',
    }

    _container_runtime_cmd = None

    pcs_resource_list = ['haproxy', 'galera', 'redis', 'ovn-dbs', 'cinder',
                         'rabbitmq', 'manila', 'ceph', 'pacemaker']

    sidecar_container_list = [
        'neutron-haproxy-ovnmeta', 'neutron-haproxy-qrouter',
        'neutron-dnsmasq-qdhcp', 'neutron-keepalived-qrouter', 'neutron-radvd']

    @property
    def container_runtime_cmd(self):
        if self._container_runtime_cmd is None:
            self._container_runtime_cmd = \
                    containers.get_container_runtime_name()
        return self._container_runtime_cmd

    @property
    def ignore_containers_list(self):
        return self.pcs_resource_list + self.sidecar_container_list

    def create_node(self, name, ssh_client, **kwargs):
        return TripleoTopologyNode(topology=self,
                                   name=name,
                                   ssh_client=ssh_client,
                                   **kwargs)

    def assert_containers_running(self, expected_containers,
                                  group=None,
                                  full_name=True, bool_check=False,
                                  nodenames=None):
        group = group or 'overcloud'
        return containers.assert_containers_running(
            group=group,
            expected_containers=expected_containers,
            full_name=full_name,
            bool_check=bool_check,
            nodenames=nodenames)

    def list_containers_td(self, group=None):
        return containers.list_containers_td(group)

    def discover_nodes(self):
        self.discover_ssh_proxy_jump_node()
        self.discover_undercloud_nodes()
        self.discover_overcloud_nodes()

    def discover_undercloud_nodes(self):
        if _undercloud.has_undercloud():
            uc_config = _undercloud.undercloud_host_config()
            ssh_client = _undercloud.undercloud_ssh_client()
            self.add_node(address=uc_config.hostname,
                          group='undercloud',
                          ssh_client=ssh_client)

    def discover_overcloud_nodes(self):
        if _undercloud.has_undercloud():
            for instance in _overcloud.list_overcloud_nodes():
                try:
                    _overcloud.power_on_overcloud_node(instance)
                except Exception:
                    LOG.exception("Error ensuring overcloud node power "
                                  "status is on")
                host_config = _overcloud.overcloud_host_config(
                    instance=instance)
                ssh_client = _overcloud.overcloud_ssh_client(
                    instance=instance,
                    host_config=host_config)
                node = self.add_node(address=host_config.hostname,
                                     group='overcloud',
                                     ssh_client=ssh_client,
                                     overcloud_instance=instance)
                assert isinstance(node, TripleoTopologyNode)
                self.discover_overcloud_node_subgroups(node)

    def discover_overcloud_node_subgroups(self, node):
        # set of subgroups extracted from node name
        subgroups: typing.Set[str] = set()

        oc_groups_dict = CONF.tobiko.tripleo.overcloud_groups_dict
        # extract subgroups names from node name
        subgroups.update(oc_groups_dict.get(subgroup) or subgroup
                         for subgroup in node.name.split('-')
                         if is_valid_overcloud_group_name(group_name=subgroup,
                                                          node_name=node.name))

        # add all those known subgroups names that are contained in
        # the node name (controller, compute, ...)
        subgroups.update(subgroup
                         for subgroup in self.known_subgroups
                         if subgroup in node.name)

        # bind node to discovered subgroups
        if subgroups:
            for subgroup in sorted(subgroups):
                LOG.debug("Add node '%s' to subgroup '%s'", node.name,
                          subgroup)
                self.add_node(hostname=node.name, group=subgroup)
        else:
            LOG.warning("Unable to obtain any node subgroup from node "
                        "name: '%s'", node.name)
        return subgroups

    def check_or_start_background_vm_ping(
            self,
            server_ip: typing.Union[str, netaddr.IPAddress],
            ssh_client: ssh.SSHClientType = None):
        if (not ssh_client and
                CONF.tobiko.tripleo.run_background_services_in_pod):
            # this fails if `oc` (openshift client) is not available
            # so, if `run_background_services_in_pod` is true, make sure
            # `oc` is available
            _openshift.check_or_start_tobiko_ping_command(server_ip)
        else:
            tripleo_nova.check_or_start_background_vm_ping(
                server_ip, ssh_client=ssh_client)

    def check_or_start_background_iperf_connection(
            self,
            server_ip: typing.Union[str, netaddr.IPAddress],
            port: int,
            protocol: str,
            ssh_client: ssh.SSHClientType = None,
            iperf3_server_ssh_client: ssh.SSHClientType = None):

        kwargs = {
            'port': port,
            'protocol': protocol,
            'ssh_client': ssh_client,
            'iperf3_server_ssh_client': iperf3_server_ssh_client,
        }

        if (not ssh_client and
                CONF.tobiko.tripleo.run_background_services_in_pod):
            # this fails if `oc` (openshift client) is not available
            # so, if `run_background_services_in_pod` is true, make sure `oc`
            # is available
            kwargs['address'] = server_ip
            kwargs['check_function'] = iperf3.check_iperf3_client_results
            kwargs['start_function'] = _openshift.start_iperf3
            kwargs['liveness_function'] = _openshift.iperf3_pod_alive
            kwargs['stop_function'] = _openshift.stop_iperf3_client
            sh.check_or_start_external_process(**kwargs)
        else:
            kwargs['server_ip'] = server_ip
            tripleo_nova.check_or_start_background_iperf_connection(**kwargs)

    def check_or_start_background_http_ping(
            self,
            server_ip: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
            ssh_client: ssh.SSHClientType = None):  # noqa; pylint: disable=W0613
        if (not ssh_client and
                CONF.tobiko.tripleo.run_background_services_in_pod):
            _openshift.check_or_start_tobiko_http_ping_command(
                server_ip=server_ip
            )
        else:
            sh.check_or_start_external_process(
                start_function=http_ping.start_http_ping_process,
                stop_function=http_ping.stop_http_ping_process,
                liveness_function=http_ping.http_ping_process_alive,
                check_function=http_ping.check_http_ping_results,
                server_ip=server_ip,
                ssh_client=ssh_client)


class TripleoTopologyNode(rhosp.RhospNode):

    def __init__(self,
                 topology: topology.OpenStackTopology,
                 name: str,
                 ssh_client: ssh.SSHClientFixture,
                 addresses: typing.Iterable[netaddr.IPAddress],
                 hostname: str,
                 overcloud_instance: metalsmith.Instance = None,
                 rhosp_version: tobiko.Version = None):
        # pylint: disable=redefined-outer-name
        super().__init__(topology=topology,
                         name=name,
                         ssh_client=ssh_client,
                         addresses=addresses,
                         hostname=hostname)
        self._overcloud_instance = overcloud_instance
        self._rhosp_version = rhosp_version

    @property
    def overcloud_instance(self) -> typing.Optional[metalsmith.Instance]:
        return self._overcloud_instance

    @property
    def rhosp_version(self) -> tobiko.Version:
        if self._rhosp_version is None:
            self._rhosp_version = self._get_rhosp_version()
        return self._rhosp_version

    l3_agent_conf_path = (
        '/var/lib/config-data/neutron/etc/neutron/l3_agent.ini')

    def list_running_servers(self) -> typing.List[nova.NovaServer]:
        running_servers = list()
        for server in nova.list_servers():
            if server.status != 'SHUTOFF':
                hypervisor_name = nova.get_server_hypervisor(server,
                                                             short=True)
                if self.name == hypervisor_name:
                    running_servers.append(server)
        return running_servers

    def power_on_node(self):
        if self.overcloud_instance is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is on...")
        _overcloud.power_on_overcloud_node(instance=self.overcloud_instance)
        hostname = sh.get_hostname(ssh_client=self.ssh_client)
        LOG.debug(f"Overcloud node {self.name} power is on ("
                  f"hostname={hostname})")

    def power_off_node(self):
        if self.overcloud_instance is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is off...")
        _overcloud.power_off_overcloud_node(instance=self.overcloud_instance)
        LOG.debug(f"Overcloud server node {self.name} power is off.")


def is_valid_overcloud_group_name(group_name: str, node_name: str = None):
    if not group_name:
        return False
    if group_name in ['overcloud', node_name]:
        return False
    if is_number(group_name):
        return False
    return True


def is_number(text: str):
    try:
        float(text)
    except ValueError:
        return False
    else:
        return True


def setup_tripleo_topology():
    if _undercloud.has_undercloud() or _overcloud.has_overcloud():
        topology.set_default_openstack_topology_class(TripleoTopology)


def actual_node_groups(groups):
    """return only existing node groups"""
    return set(groups).intersection(topology.list_openstack_node_groups())


def get_node(node_name):
    node_name = node_name.split('.')[0]
    return [node for node in topology.list_openstack_nodes() if
            node.name == node_name][0]


def is_composable_roles_env():
    composable_nodes = ['messaging', 'database', 'networker']
    for nodes in composable_nodes:
        if nodes in topology.list_openstack_node_groups():
            return True
    return False



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr

from tobiko.shell import ss
from tobiko.tests import unit


class SockHeaderTest(unit.TobikoUnitTest):

    def test_parse_header_basic(self):
        header_str = 'Recv-Q Send-Q Local Address:Port Peer Address:Port'
        header = ss.SockHeader(header_str)
        self.assertEqual(['recv_q', 'send_q', 'local', 'remote'],
                         list(header))

    def test_parse_header_with_netid(self):
        header_str = 'Netid Recv-Q Send-Q Local Address:Port Peer Address:Port'
        header = ss.SockHeader(header_str)
        self.assertEqual(['protocol', 'recv_q', 'send_q', 'local', 'remote'],
                         list(header))

    def test_parse_header_with_state(self):
        header_str = 'State Recv-Q Send-Q Local Address:Port Peer Address:Port'
        header = ss.SockHeader(header_str)
        self.assertEqual(['state', 'recv_q', 'send_q', 'local', 'remote'],
                         list(header))

    def test_parse_header_with_process(self):
        header_str = ('Recv-Q Send-Q Local Address:Port '
                      'Peer Address:Port Process')
        header = ss.SockHeader(header_str)
        self.assertEqual(['recv_q', 'send_q', 'local', 'remote', 'process'],
                         list(header))

    def test_extend_ports(self):
        header_str = 'Recv-Q Send-Q Local Address:Port Peer Address:Port'
        header = ss.SockHeader(header_str)
        header.extend_ports()
        self.assertEqual(['recv_q', 'send_q', 'local_addr', 'local_port',
                          'remote_addr', 'remote_port'],
                         list(header))


class GetProcessesTest(unit.TobikoUnitTest):

    def test_single_process(self):
        processes_str = 'users:(("httpd",pid=735448,fd=11))'
        result = ss.get_processes(processes_str)
        self.assertEqual(['httpd'], result)

    def test_multiple_processes(self):
        """Test parsing multiple processes

        Note: The current implementation of get_processes only returns the
        first process name from the list. This appears to be a limitation
        of the current implementation.
        """
        processes_str = ('users:(("httpd",pid=4969,fd=53),'
                         '("httpd",pid=3328,fd=53))')
        result = ss.get_processes(processes_str)
        # Current behavior returns only first process
        self.assertEqual(['httpd'], result)

    def test_different_process_names(self):
        """Test parsing different process names

        Note: The current implementation of get_processes only returns the
        first process name from the list. This appears to be a limitation
        of the current implementation.
        """
        processes_str = ('users:(("nginx",pid=1234,fd=10),'
                         '("worker",pid=5678,fd=20))')
        result = ss.get_processes(processes_str)
        # Current behavior returns only first process
        self.assertEqual(['nginx'], result)


class ParseTcpSocketTest(unit.TobikoUnitTest):

    def test_parse_ipv4_socket(self):
        """Test parsing IPv4 socket with specific addresses"""
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     192.168.1.1:80 192.168.1.2:54321 '
            'users:(("httpd",pid=1234,fd=5))')

        result = ss.parse_tcp_socket(headers, sock_info)

        self.assertEqual('0', result['recv_q'])
        self.assertEqual('10', result['send_q'])
        self.assertEqual(netaddr.IPAddress('192.168.1.1'),
                         result['local_addr'])
        self.assertEqual('80', result['local_port'])
        self.assertEqual(netaddr.IPAddress('192.168.1.2'),
                         result['remote_addr'])
        self.assertEqual('54321', result['remote_port'])
        self.assertEqual(['httpd'], result['process'])

    def test_parse_ipv4_wildcard_both_addresses(self):
        """Test parsing IPv4 socket with wildcard 0.0.0.0 addresses"""
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     0.0.0.0:6641 0.0.0.0:* '
            'users:(("ovsdb-server",pid=62494,fd=29))')

        result = ss.parse_tcp_socket(headers, sock_info)

        self.assertEqual(netaddr.IPAddress('0.0.0.0'), result['local_addr'])
        self.assertEqual('6641', result['local_port'])
        self.assertEqual(netaddr.IPAddress('0.0.0.0'), result['remote_addr'])
        self.assertEqual('*', result['remote_port'])
        self.assertEqual(['ovsdb-server'], result['process'])

    def test_parse_ipv4_wildcard_asterisk(self):
        """Test parsing IPv4 socket with asterisk wildcard (listening)

        This tests the fix for netaddr 1.3.0 compatibility where '*' should
        be converted to '0.0.0.0' for IPv4 sockets.
        """
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     192.168.1.1:80 *:* '
            'users:(("httpd",pid=1234,fd=5))')

        result = ss.parse_tcp_socket(headers, sock_info)

        # When local is a specific IPv4, wildcard remote should be 0.0.0.0
        self.assertEqual(netaddr.IPAddress('192.168.1.1'),
                         result['local_addr'])
        self.assertEqual('80', result['local_port'])
        self.assertEqual(netaddr.IPAddress('0.0.0.0'), result['remote_addr'])
        self.assertEqual('*', result['remote_port'])

    def test_parse_ipv6_socket(self):
        """Test parsing IPv6 socket with specific addresses"""
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     [fd00:192:170:1::166]:80 '
            '[fd00:192:170:1::167]:54321 users:(("httpd",pid=1234,fd=5))')

        result = ss.parse_tcp_socket(headers, sock_info)

        self.assertEqual(netaddr.IPAddress('fd00:192:170:1::166'),
                         result['local_addr'])
        self.assertEqual('80', result['local_port'])
        self.assertEqual(netaddr.IPAddress('fd00:192:170:1::167'),
                         result['remote_addr'])
        self.assertEqual('54321', result['remote_port'])

    def test_parse_ipv6_wildcard_asterisk(self):
        """Test parsing IPv6 socket with asterisk wildcard (listening)

        This tests the fix for netaddr 1.3.0 compatibility where '*' should
        be converted to '::' for IPv6 sockets. This is the actual case that
        was failing in IPv6 environments.
        """
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     *:6641 *:* '
            'users:(("ovsdb-server",pid=61929,fd=29))')

        result = ss.parse_tcp_socket(headers, sock_info)

        # When both are wildcards and no IP version hint, defaults to IPv6
        self.assertEqual(netaddr.IPAddress('::'), result['local_addr'])
        self.assertEqual('6641', result['local_port'])
        self.assertEqual(netaddr.IPAddress('::'), result['remote_addr'])
        self.assertEqual('*', result['remote_port'])
        self.assertEqual(['ovsdb-server'], result['process'])

    def test_parse_ipv6_wildcard_with_specific_local(self):
        """Test parsing IPv6 socket with specific local and wildcard remote"""
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     [fd00:192:170:1::166]:6641 *:* '
            'users:(("ovsdb-server",pid=61929,fd=29))')

        result = ss.parse_tcp_socket(headers, sock_info)

        # When local is IPv6, wildcard remote should be ::
        self.assertEqual(netaddr.IPAddress('fd00:192:170:1::166'),
                         result['local_addr'])
        self.assertEqual('6641', result['local_port'])
        self.assertEqual(netaddr.IPAddress('::'), result['remote_addr'])
        self.assertEqual('*', result['remote_port'])

    def test_parse_socket_with_state(self):
        """Test parsing socket with state field"""
        headers = ss.SockHeader(
            'State Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            'ESTAB 0      0     192.168.1.1:22 192.168.1.2:54321 '
            'users:(("sshd",pid=1234,fd=5))')

        result = ss.parse_tcp_socket(headers, sock_info)

        self.assertEqual('ESTAB', result['state'])
        self.assertEqual('0', result['recv_q'])
        self.assertEqual('0', result['send_q'])

    def test_parse_socket_multiple_processes(self):
        """Test parsing socket with multiple processes

        Note: Due to current limitation of get_processes, only the first
        process name is captured.
        """
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine(
            '0      10     192.168.1.1:80 192.168.1.2:54321 '
            'users:(("httpd",pid=4969,fd=53),("httpd",pid=3328,fd=53))')

        result = ss.parse_tcp_socket(headers, sock_info)

        # Current behavior captures only first process
        self.assertEqual(['httpd'], result['process'])

    def test_parse_socket_invalid_line(self):
        """Test parsing fails gracefully with invalid line"""
        headers = ss.SockHeader(
            'Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        sock_info = ss.SockLine('0 10 192.168.1.1:80')  # Missing fields

        ex = self.assertRaises(ValueError,
                               ss.parse_tcp_socket,
                               headers,
                               sock_info)
        self.assertIn('Unable to parse line', str(ex))


class ParseUnixSocketTest(unit.TobikoUnitTest):

    def test_parse_unix_socket(self):
        """Test parsing Unix domain socket"""
        headers = ss.SockHeader(
            'Netid Recv-Q Send-Q Local Address:Port Peer Address:Port Process')
        headers.extend_ports()
        sock_info = ss.SockLine(
            'u_str 0      0     /var/run/socket.sock * * * '
            'users:(("daemon",pid=1234,fd=5))')

        result = ss.parse_unix_socket(headers, sock_info)

        self.assertEqual('u_str', result['protocol'])
        self.assertEqual('0', result['recv_q'])
        self.assertEqual('0', result['send_q'])
        self.assertEqual('/var/run/socket.sock', result['local_addr'])
        self.assertEqual(['daemon'], result['process'])



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
from oslo_log import log as logging

from tobiko.shell import custom_script
from tobiko.shell import files
from tobiko.shell import ssh


LOG = logging.getLogger(__name__)

TIMEOUT = 2  # seconds
LOG_FILE_NAME = "dns_ping.log"
DNS_PING_SCRIPT_NAME = "tobiko_dns_ping.sh"
DNS_PING_SCRIPT = """
#!/bin/bash

HOSTNAME_CMD=$(which hostname)
DATE_CMD=$(which date)
RESOLVE_CMD="$(which resolvectl) query --cache=no --legend=no --stale-data=no"

ip_address=$1;
fqdn=$2
output_file=$3;

rm $output_file;

# First get IP address(es) of the machine
server_ips=$($HOSTNAME_CMD --all-ip-addresses)

while true; do
    current_time=$(/usr/bin/date +"{time_format}");
    $RESOLVE_CMD $fqdn | grep -q $ip_address

    if [ $? -eq 0 ]; then
            response="{result_ok}";
    else
            response="{result_failed}";
    fi;
    echo "{output_result_line}" >> $output_file;
    sleep {timeout};
done;
""".format(  # noqa: E501
    time_format=custom_script.LOG_TIME_FORMAT,
    result_ok=custom_script.RESULT_OK,
    result_failed=custom_script.RESULT_FAILED,
    output_result_line=custom_script.LOG_RESULT_FORMAT,
    timeout=TIMEOUT,
)


def _ensure_script_is_on_server(ssh_client: ssh.SSHClientType) -> None:
    custom_script.ensure_script_is_on_server(
        DNS_PING_SCRIPT_NAME,
        DNS_PING_SCRIPT,
        ssh_client=ssh_client)


def _get_script_command(
        ip_address: typing.Union[str, netaddr.IPAddress],
        fqdn: str,
        ssh_client: ssh.SSHClientType) -> str:
    homedir = files.get_homedir(ssh_client)
    logfile_path = _get_logfile_path(ssh_client)
    return (f"bash {homedir}/{DNS_PING_SCRIPT_NAME} "
            f"{ip_address} {fqdn} {logfile_path}")


def _get_log_dir(ssh_client: ssh.SSHClientType = None) -> str:
    return custom_script.get_log_dir(
        "tobiko_dns_ping_results", ssh_client)


def _get_logfile_path(ssh_client: ssh.SSHClientType) -> str:
    return f"{_get_log_dir(ssh_client)}/{LOG_FILE_NAME}"


def _get_dns_ping_pid(
        ip_address: typing.Union[str, netaddr.IPAddress],
        fqdn: str,
        ssh_client: ssh.SSHClientType) -> typing.Union[int, None]:
    processes = custom_script.get_process_pid(
        command_line=_get_script_command(ip_address, fqdn, ssh_client),
        ssh_client=ssh_client)
    if not processes:
        LOG.debug('no DNS ping script found.')
    return processes


def start_dns_ping_process(
        ip_address: typing.Union[str, netaddr.IPAddress],
        fqdn: str,
        ssh_client: ssh.SSHClientType) -> None:
    _ensure_script_is_on_server(ssh_client)
    if dns_ping_process_alive(ip_address, fqdn, ssh_client):
        return
    custom_script.start_script(
        _get_script_command(ip_address, fqdn, ssh_client),
        ssh_client=ssh_client)


def stop_dns_ping_process(
        ip_address: typing.Union[str, netaddr.IPAddress],
        fqdn: str,
        ssh_client: ssh.SSHClientType) -> None:
    pid = _get_dns_ping_pid(ip_address, fqdn, ssh_client)
    if pid:
        custom_script.stop_script(pid, ssh_client=ssh_client)


def dns_ping_process_alive(
        ip_address: typing.Union[str, netaddr.IPAddress],
        fqdn: str,
        ssh_client: ssh.SSHClientType) -> bool:
    return bool(_get_dns_ping_pid(ip_address, fqdn, ssh_client))


def check_dns_ping_results(
        ssh_client: ssh.SSHClientType,  # noqa; pylint: disable=W0613
        **kwargs) -> None:
    # Source log file is on the guest vm so ssh_client needs to be used
    # to get it
    src_logfile = _get_logfile_path(ssh_client)
    # Destination is local to where Tobiko is running so no need to pass
    # ssh_client to get_log_dir() function this time
    dest_logfile = f"{_get_log_dir()}/{LOG_FILE_NAME}"
    custom_script.copy_log_file(src_logfile, dest_logfile, ssh_client)

    # Looking for the files locally so no need to pass ssh_client to the
    # _get_log_dir function
    logfiles = custom_script.get_log_files(
        glob_log_pattern=f"{_get_log_dir()}/{LOG_FILE_NAME}")

    custom_script.check_results(logfiles)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import time

from cliff import command
from oslo_log import log as logging
from oslo_serialization import jsonutils

from tobiko.shell import http_ping


LOG = logging.getLogger(__name__)


class TobikoHttpPing(command.Command):

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server_ip',
            help='IP address of the server to send requests to'
        )
        parser.add_argument(
            '--interval',
            type=int,
            default=5,
            help='Interval of the HTTP requests.'
        )
        return parser

    def take_action(self, parsed_args):
        try:
            LOG.debug("Starting sending HTTP requests to the server: %s",
                      parsed_args.server_ip)
            output_filename = f'http_ping_{parsed_args.server_ip}.log'
            result_file_path = \
                f"{http_ping.get_log_dir()}/{output_filename}"
            while True:
                with open(result_file_path, "a") as result_file:
                    iteration_start_time = time.time()
                    ping_result = http_ping.http_ping(parsed_args.server_ip)
                    result_file.write(jsonutils.dumps(ping_result) + "\n")
                    LOG.info("%s response from %s %s",
                             ping_result['time'], parsed_args.server_ip,
                             ping_result['response'])
                    elapsed = time.time() - iteration_start_time
                    if elapsed < parsed_args.interval:
                        time.sleep(parsed_args.interval - elapsed)
        except Exception as e:
            LOG.error("Failed to send http request to the server '%s'. "
                      "Error: %s", parsed_args.server_ip, e)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.custom_script import _constants
from tobiko.shell.custom_script import _custom_script


RESULT_OK = _constants.RESULT_OK
RESULT_FAILED = _constants.RESULT_FAILED
LOG_TIME_FORMAT = _constants.LOG_TIME_FORMAT
LOG_RESULT_FORMAT = _constants.LOG_RESULT_FORMAT

ensure_script_is_on_server = _custom_script.ensure_script_is_on_server
get_log_dir = _custom_script.get_log_dir
get_log_files = _custom_script.get_log_files
copy_log_file = _custom_script.copy_log_file
get_process_pid = _custom_script.get_process_pid
get_process_pids = _custom_script.get_process_pids
check_results = _custom_script.check_results

start_script = _custom_script.start_script
stop_script = _custom_script.stop_script



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


import os

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


def get_homedir(ssh_client: ssh.SSHClientType = None) -> str:
    if ssh_client:
        return sh.execute(
            'echo ~', ssh_client=ssh_client).stdout.rstrip()
    else:
        return sh.get_user_home_dir()


def get_home_absolute_filepath(path: str,
                               ssh_client: ssh.SSHClientType = None) -> str:
    if ssh_client is None:
        return _get_local_filepath(path)
    else:
        return _get_remote_filepath(path, ssh_client)


def _get_local_filepath(path: str) -> str:
    final_dir_path = f'{sh.get_user_home_dir()}/{path}'
    if not os.path.exists(final_dir_path):
        os.makedirs(final_dir_path)
    return final_dir_path


def _get_remote_filepath(path: str,
                         ssh_client: ssh.SSHClientType) -> str:
    homedir = get_homedir(ssh_client)
    final_dir_path = f'{homedir}/{path}'
    sh.make_remote_dirs(final_dir_path, ssh_client=ssh_client)
    return final_dir_path


def truncate_client_logfile(
        logfile: str,
        ssh_client: ssh.SSHClientType = None) -> None:
    if ssh_client:
        _truncate_remote_logfile(logfile, ssh_client)
    else:
        tobiko.truncate_logfile(logfile)


def _truncate_remote_logfile(logfile: str,
                             ssh_client: ssh.SSHClientType) -> None:
    truncated_logfile = tobiko.get_truncated_filename(logfile)
    sh.execute(f'/usr/bin/mv {logfile} {truncated_logfile}',
               ssh_client=ssh_client)


def remove_old_logfile(logfile: str,
                       ssh_client: ssh.SSHClientType = None):
    if ssh_client:
        sh.execute(f'/usr/bin/rm -f {logfile}',
                   ssh_client=ssh_client)
    else:
        try:
            os.remove(logfile)
        except FileNotFoundError:
            pass



# Copyright (c) 2024 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import sys

from cliff import app
from cliff import commandmanager
from oslo_log import log as logging
from pbr import version


class Main(app.App):

    log = logging.getLogger(__name__)

    def __init__(self):
        super(Main, self).__init__(
            description='Tobiko CLI application',
            version=version.VersionInfo('tobiko').version_string_with_vcs(),
            command_manager=commandmanager.CommandManager(
                'tobiko.cli_commands'),
            deferred_help=True,
            )

    def initialize_app(self, argv):
        self.log.debug('tobiko initialize_app')

    def prepare_to_run_command(self, cmd):
        self.log.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        self.log.debug('tobiko clean_up %s', cmd.__class__.__name__)
        if err:
            self.log.debug('tobiko got an error: %s', err)


def main(argv=sys.argv[1:]):
    the_app = Main()
    return the_app.run(argv)



# Copyright (c) 2024 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import sys

from cliff import command
from oslo_log import log as logging

from tobiko.shell import ping

LOG = logging.getLogger(__name__)


class TobikoPing(command.Command):

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            help='Address of the server to ping'
        )
        parser.add_argument(
            '-i', '--interval',
            default=None,
            help="Seconds of time interval between "
                 "consecutive before ICMP messages"
        )
        parser.add_argument(
            '--result-file',
            default='tobiko_ping_results',
            help='Name of the directory where ping log files are '
                 'stored.'
        )
        return parser

    def take_action(self, parsed_args):
        error_code = 0
        try:
            LOG.debug("Starting ping server: %s", parsed_args.server)
            ping.write_ping_to_file(ping_ip=parsed_args.server,
                                    output_dir=parsed_args.result_file,
                                    interval=parsed_args.interval)
            LOG.debug("Finished ping server: %s", parsed_args.server)
        except Exception as e:
            if hasattr(e, 'errno'):
                error_code = e.errno
            else:
                error_code = 1
            LOG.error("Failed to ping server '%s'. Error: %s",
                      parsed_args.server, e)
        if error_code:
            sys.exit(error_code)
